"""Module for managing env variables."""

from functools import cache

from pydantic import BaseModel
from pydantic_settings import BaseSettings, SettingsConfigDict


class DbStorage(BaseModel):
    """Settings for database."""

    drivername: str = ""
    host: str = ""
    port: int = -1
    database: str = ""
    user: str = ""
    password: str = ""


class Redis(BaseModel):
    """Settings for redis."""

    host: str = ""
    port: int = -1
    password: str = ""
    queue_db: int = -1


class Celery(BaseModel):
    """Setting for celery."""

    simulator_aer_queue: str = ""
    compiler_qiskit_queue: str = ""
    compiler_zne_pass_queue: str = ""
    margin_after_soft_timeout: float = 1.0


class EnvironmentSettings(BaseSettings):
    """Class holder all different env settings."""

    db: DbStorage = DbStorage()
    redis: Redis = Redis()
    celery: Celery = Celery()

    model_config = SettingsConfigDict(
        env_file_encoding="utf-8",
        env_nested_delimiter="__",
    )


@cache
def get_settings():
    """Returns cached version of env settings."""
    return EnvironmentSettings()



"""Manages communication with Keycloak server."""
import time

from keycloak import KeycloakOpenID

from .exceptions import KeycloakTokenManagerError


class KeycloakTokenManager:
    """Manages communication with Keycloak server.

    It checks whether the token is still valid and handles
    getting and refreshing tokens when needed.
    """

    def __init__(self, server_address: str, username: str, password: str):
        """Initialize the KeycloakTokenManager with server address, username, and password."""
        self.server_url = server_address.rstrip("/") + "/auth/"
        self.username = username
        self.password = password

        self.token = None
        self.header = None
        self.token_request_timestamp = 0.0
        self.token_expiration_time = 0
        self.token_refresh_expiration_time = 0

        self.keycloak_openid = KeycloakOpenID(
            server_url=self.server_url,
            client_id="react-client",
            realm_name="qciconnect",
            verify=True,
        )

    def _update_timestamps(self) -> None:
        """Update the timestamp for token request and store information concerning expiration."""
        self.token_request_timestamp = time.time()
        self.token_expiration_time = int(self.token["expires_in"])
        self.token_refresh_expiration_time = int(self.token["refresh_expires_in"])

    def _is_token_expired(self) -> bool:
        """Check whether the current token has expired."""
        return self.token_request_timestamp + self.token_expiration_time < time.time()

    def _is_token_refreshable(self) -> bool:
        """Check whether the current token can be refreshed."""
        return not (
            self._is_token_expired()
            and self.token_request_timestamp + self.token_refresh_expiration_time < time.time()
        )

    def _ensure_token_validity(self) -> None:
        """Update the token, if necessary."""
        if self.token is None:
            try:
                self.token = self.keycloak_openid.token(
                    username=self.username, password=self.password
                )
                self._update_timestamps()
            except Exception as e:
                raise KeycloakTokenManagerError(f"Error obtaining token from Keycloak: {e}") from e

        if self._is_token_expired() and self._is_token_refreshable():
            try:
                self.token = self.keycloak_openid.refresh_token(
                    refresh_token=self.token["refresh_token"]
                )
                self._update_timestamps()
            except Exception as e:
                raise KeycloakTokenManagerError(f"Error refreshing token from Keycloak: {e}") from e

        if self._is_token_expired() and not self._is_token_refreshable():
            try:
                self.token = self.keycloak_openid.token(
                    username=self.username, password=self.password
                )
                self._update_timestamps()
            except Exception as e:
                raise KeycloakTokenManagerError(f"Error obtaining token from Keycloak: {e}") from e

    def get_auth_header(self) -> dict:
        """Get a dictionary that comprises the authorization header for the API.

        Returns: A dictionary with an authorization token.
        """
        self._ensure_token_validity()
        return {"Authorization": f"Bearer {self.token['access_token']}"}

    def add_auth_header(self, headers: dict) -> dict:
        """Add the "Authorization" header to a dictionary of headers.

        If necessary, a new token is created (or an existing one is refreshed).

        Args:
            headers (dict): A dictionary of headers.

        Returns:
            dict: The updated input dictionary with the "Authorization" header added.
        """
        auth_header = self.get_auth_header()
        headers.update(auth_header)
        return headers


class DummyTokenManager:
    """Dummy token manager for use with mini-orchestrator."""
    def __init__(self):
        """Init Dummy token manager for use with mini-orchestrator."""
        pass

    def add_auth_header(self, headers: dict) -> dict:
        """Returns unmodified headers dictionary."""
        return headers



"""QCI Connect Logging configuration."""

import logging
import os
from logging import config

from qciconnect_logging.logger import (
    DEFAULT_LOG_MESSAGE_FORMAT,
    DEFAULT_LOG_MESSAGE_FORMAT_JSON,
    DEFAULT_LOGGER_NAME,
)

STREAM_HANDLER_CLASS = "logging.StreamHandler"
FILE_HANDLER_CLASS = "logging.FileHandler"
STDOUT_STREAM = "ext://sys.stdout"
LOG_FILENAME = os.getenv("LOG_FILENAME", "logs.log")
LOG_PATH = f"/app/logs/{LOG_FILENAME}"
LOG_LEVEL = os.getenv("LOG_LEVEL", "DEBUG")
DEV_ENV = os.getenv("DEV_ENV", "true").lower() == "true"

class SingleLineFormatter(logging.Formatter):
    """Custom formatter to replace newline characters with spaces in log messages."""

    def format(self, record) -> str:
        """Format log record."""
        if isinstance(record.msg, str):
            record.msg = record.msg.replace('\n', ' ')
            record.msg = record.msg.replace('\r', ' ')
            record.msg = record.msg.replace('\t', ' ')
            record.msg = record.msg.replace('"', '\\"')
        return super().format(record)

if DEV_ENV:
    message_format = DEFAULT_LOG_MESSAGE_FORMAT
    handlers = {
        "stdout_default": {
            "formatter": "formatter_default",
            "class": STREAM_HANDLER_CLASS,
            "stream": STDOUT_STREAM,
            "level": LOG_LEVEL,
        },
    }
    celery_log_level = "INFO"
    sqlalchemy_log_level = "INFO"
else:
    message_format = DEFAULT_LOG_MESSAGE_FORMAT_JSON
    handlers = {
        "stdout_default": {
            "formatter": "formatter_default",
            "class": STREAM_HANDLER_CLASS,
            "stream": STDOUT_STREAM,
            "level": LOG_LEVEL,
        },
    } 
    celery_log_level = "WARNING"
    sqlalchemy_log_level = "WARNING"

log_config = {
    "version": 1,
    "formatters": {
        "formatter_default": {"()": SingleLineFormatter,"format":message_format},
    },
    "handlers": handlers,
    "loggers": {
        # root logger to stdout
        "": {"handlers": ["stdout_default"], "level": LOG_LEVEL},
        # QCI Connect  default technical logger
        DEFAULT_LOGGER_NAME: {
            "handlers": ["stdout_default"],
            "level": LOG_LEVEL,
            "propagate": False,
        },
        # 3 uvicorn loggers
        "uvicorn.access": {
            "handlers": ["stdout_default"],
            "level": "INFO",
            "propagate": False,
        },
        "uvicorn.error": {
            "handlers": ["stdout_default"],
            "level": "INFO",
            "propagate": False,
        },
        "uvicorn.asgi": {
            "handlers": ["stdout_default"],
            "level": "INFO",
            "propagate": False,
        },
        # python warnings propagated to stdout
        "py.warnings": {
            "handlers": ["stdout_default"],
            "level": "INFO",
            "propagate": False,
        },
        # SQL queries to stdout
        "sqlalchemy.engine": {
            "handlers": ["stdout_default"],
            "level": sqlalchemy_log_level,
            "propagate": False,
        },
        # celery logging to stdout
        "celery": {"handlers": ["stdout_default"], "level": celery_log_level, "propagate": False},
    },
}


def initialize_logging():
    """Initialize logging configuration based on dict config."""
    # capture warnings with logger
    logging.captureWarnings(True)
    # configure logging from dict configuration
    config.dictConfig(log_config)



"""Technical logger for QCI Connect application."""

import logging
from json import dumps

# default stdout logging format
DEFAULT_LOG_MESSAGE_FORMAT_JSON = dumps({
    "time": "%(asctime)s",
    "processName": "%(processName)s",
    "PID": "%(process)d",
    "threadName": "%(threadName)s",
    "TID": "%(thread)d",
    "module": "%(module)s",
    "funcName": "%(funcName)s",
    "name": "%(name)s",
    "level": "%(levelname)s",
    "log": "%(message)s"
})

DEFAULT_LOG_MESSAGE_FORMAT = (
    "%(asctime)s - %(processName)s (PID %(process)d): %(threadName)s "
    "(TID %(thread)d) - %(module)s - %(funcName)s - %(name)s - %(levelname)s "
    "- %(message)s"
)


# name of the default technical logger
DEFAULT_LOGGER_NAME = "qciconnect.default"


class QCIConnectLogging:
    """Default technical logger."""

    @classmethod
    def get_logger(cls):
        """Get default logger of the application."""
        return logging.getLogger(DEFAULT_LOGGER_NAME)



"""Exception classes for qciconnect package."""

class KeycloakTokenManagerError(Exception):
    """Exception class for errors from KeycloakTokenManager."""


class QciConnectClientError(Exception):
    """Exception class for errors from QciConnectClient."""


class InvalidIdentifierError(Exception):
    """Thrown when identifiers contain forbidden characters."""


class ResultError(Exception):
    """Exception class for errors related to results handling."""



"""Module that provides celery instances and configs."""

from celery import Celery
from qciconnect_settings.environment_config import get_settings

# Fetch settings
settings = get_settings()
_db = settings.db
_redis = settings.redis


def _get_redis_url_template() -> str:
    """Return redis url template.

    Returns:
        redis url template string.
    """
    return "redis://:{password}@{host}:{port}/{db}"


def _get_postgres_url_template() -> str:
    """Return postgres url template.

    Returns:
        postres url template string.
    """
    return "db+{drivername}://{user}:{password}@{host}/{database_name}"


# default redis and postgres url for celery
redis_url = _get_redis_url_template().format(
    password=_redis.password, host=_redis.host, port=_redis.port, db=_redis.queue_db
)
postgres_url = _get_postgres_url_template().format(
    drivername=_db.drivername,
    user=_db.user,
    password=_db.password,
    host=_db.host,
    database_name=_db.database,
)


def get_celery_app(
    name: str,
    broker_url: str = redis_url,
    backend_url: str = postgres_url,
    task_routes: dict | None = None,
    registered_task_modules: list[str] | None = None,
    task_acks_late=False,
    worker_prefetch_multiplier=1,
    task_track_started=True,
) -> Celery:
    """Create a new Celery application.

    Args:
        name (str): The name of the celery application.
        broker_url (str): The URL of the message broker to use for task distribution.
        backend_url (str): The URL of the backend to use for storing task results.
        task_routes (dict): Dict with task queue routing information. Defaults to empty dict.
        registered_task_modules (list[str], optional): A list of Python modules that
            register tasks to be used by the Celery application. Defaults to None.
        task_acks_late (bool, optional): Configuration option to acknowledge tasks after they have
            been successfully processed by the worker instead of immediately after receipt.
            Defaults to False.
        worker_prefetch_multiplier (int, optional): The number of tasks a worker will prefetch
            from the task queue. Defaults to 1.
        task_track_started (bool, optional): If True, a 'STARTED' state is sent when
            the worker starts processing the task. Defaults to True.

    Returns:
        celery.Celery: Configured Celery application instance.
    """
    if task_routes is None:
        task_routes = {}

    app = Celery(
        name,
        backend=backend_url,
        broker=broker_url,
        task_routes=task_routes,
        imports=registered_task_modules,
        task_acks_late=task_acks_late,
        worker_prefetch_multiplier=worker_prefetch_multiplier,
        task_track_started=task_track_started,
    )
    app.conf.update(worker_hijack_root_logger=False)
    return app
